// Copyright 2024 schukai GmbH
// SPDX-License-Identifier: proprietary

package data

import (
	"testing"
)

func TestNewTransformer(t *testing.T) {
	tt := NewTransformer(&map[any]any{})
	if tt == nil {
		t.Error("Transformer is nil")
	}
}

func TestTransformValues(t *testing.T) {

	demo := map[any]any{
		0: map[string]any{
			"name":         "Hannes",
			"plz":          "12345",
			"space-around": "  12345  ",
			"comment":      "<h1>Test</h1>",
			"url":          "$ & < > ? ; # : = , \" ' ~ + % { } | \\ ^ [ ] `",
			"url-decoded":  "%24+%26+%3C+%3E+%3F+%3B+%23+%3A+%3D+%2C+%22+%27+~+%2B+%25+%7B+%7D+%7C+%5C+%5E+%5B+%5D+%60",
		},
		1: map[string]any{
			"float":  1.2345,
			"int":    12345,
			"bool":   true,
			"string": "12345",
			"number": "12345",
		},
		2: map[string]any{
			"json": "{\"bool\":true,\"float\":1.2345,\"int\":12345,\"number\":\"12345\",\"string\":\"12345\"}",
		},
	}

	testData := []struct {
		instruction string
		expected    any
		data        map[any]any
		hasError    bool
	}{

		{
			instruction: `static:"This is a test" | split:" " | index:1`,
			expected:    "is",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"This is a test" | replace:"test":"test2"`,
			expected:    "This is a test2",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:65 | int | char`,
			expected:    "A",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:60 | int | add`,
			expected:    nil,
			data:        nil,
			hasError:    true,
		},
		{
			instruction: `static:60 | int | add:6 | sub:1 | char`,
			expected:    "A",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:65 | int | mul:2 | div:2 | char`,
			expected:    "A",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:65 | int | mod:2`,
			expected:    int64(1),
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:66 | int | mod:2`,
			expected:    int64(0),
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:65.5123423 | float | round:3`,
			expected:    65.512,
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:65.5 | float | floor`,
			expected:    65.0,
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:65.5 | float | ceil`,
			expected:    66.0,
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"this is a test" | ucfirst`,
			expected:    "This is a test",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"this is a test" | ucwords`,
			expected:    "This Is A Test",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"Iñtërnâtiônàlizætiøn" | len`,
			expected:    27,
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"Iñtërnâtiônàlizætiøn" | base64`,
			expected:    "ScOxdMOrcm7DonRpw7Ruw6BsaXrDpnRpw7hu",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"Iñtërnâtiônàlizætiøn" | base64 | base64decode`,
			expected:    "Iñtërnâtiônàlizætiøn",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"<a href='test'>Test</a>" | htmlspecialchars`,
			expected:    "&lt;a href=&#39;test&#39;&gt;Test&lt;/a&gt;",
			data:        map[any]any{},
			hasError:    false,
		}, {
			instruction: `static:"xytz" | md5`,
			expected:    "328ac2a6c0c449473275c7a029ae8cca",
			data:        map[any]any{},
			hasError:    false,
		}, {
			instruction: `static:"xytz" | sha1`,
			expected:    "a17df824687961eb92922d43f76db3a0444489c5",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"xytz" | strpad:10:"*":left`,
			expected:    "******xytz",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"xytz" | strpad:10:"*":right`,
			expected:    "xytz******",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"xytz" | strpad:10:"*":both`,
			expected:    "***xytz***",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"xytz" | strpad:9:"*":both`,
			expected:    "**xytz***",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"xytz" | reverse`,
			expected:    "ztyx",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"1234567890" | substr:4:3`,
			expected:    "567",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"1234567890" | substr:0`,
			expected:    "1234567890",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"1234567890" | substr:0:0`,
			expected:    "",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"1234567890" | substr:1:1`,
			expected:    "2",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"1234567890" | substr:1:100`,
			expected:    "234567890",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"1234567890" | substr:1`,
			expected:    "234567890",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"1234567890" | substr:9`,
			expected:    "0",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"1234567890" | substr:20`,
			expected:    "",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"xytz" | strrepeat:3`,
			expected:    "xytzxytzxytz",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"1234 567 8901 2345 67890 12345 67 890" | wordwrap:10:"|"`,
			expected:    "1234 567|8901 2345|67890|12345 67|890",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"xytz" | sha256`,
			expected:    "e8e8e8e492039a541a9cb9a31d752e31849d26529dc8d7bafcda20741fb1708c",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"ScOxdMOrcm7DonRpw7Ruw6BsaXrDpnRpw7hu" | base64decode`,
			expected:    "Iñtërnâtiônàlizætiøn",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"test" | len`,
			expected:    4,
			data:        map[any]any{},
			hasError:    false,
		},

		{
			instruction: `static:"test" | prefix:"pre"`,
			expected:    "pretest",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:"test" | suffix:"suf"`,
			expected:    "testsuf",
			data:        map[any]any{},
			hasError:    false,
		},

		////// JSON
		{
			instruction: `dataset:entry | json`,
			expected:    "{\"bool\":true,\"float\":1.2345,\"int\":12345,\"number\":\"12345\",\"string\":\"12345\"}",
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},
		{
			instruction: `static:"1" | json`,
			expected:    "\"1\"",
			data:        nil,
			hasError:    false,
		},
		{
			instruction: `static:"1" | fromjson`,
			expected:    1.0,
			data:        nil,
			hasError:    false,
		},
		{
			instruction: `static:true | fromjson`,
			expected:    true,
			data:        nil,
			hasError:    false,
		},
		{
			instruction: `static:null | fromjson`,
			expected:    nil,
			data:        nil,
			hasError:    false,
		},
		{
			instruction: `static:"1" |int | json`,
			expected:    "1", // json always returns a string
			data:        nil,
			hasError:    false,
		}, {
			instruction: `static:"1" |bool | json`,
			expected:    "true", // json always returns a string
			data:        nil,
			hasError:    false,
		}, {
			instruction: `static:"1" |float | json`,
			expected:    "1", // json always returns a string
			data:        nil,
			hasError:    false,
		},
		{
			instruction: `dataset:entry | index:json | fromjson | tojson`,
			expected:    "{\"bool\":true,\"float\":1.2345,\"int\":12345,\"number\":\"12345\",\"string\":\"12345\"}",
			data: map[any]any{
				"entry": demo[2],
			},
			hasError: false,
		},

		////// FLOAT TO
		{
			instruction: `dataset:entry | index:"float" | tostring`,
			expected:    "1.2345",
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:"float" | toboolean`,
			expected:    true,
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:"float" | toint`,
			expected:    1,
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:"float" | tofloat`,
			expected:    1.2345,
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},

		//// INT TO

		{
			instruction: `dataset:entry | index:"int" | tostring`,
			expected:    "12345",
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:"int" | toboolean`,
			expected:    true,
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:"int" | toint`,
			expected:    int64(12345),
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:"int" | tofloat`,
			expected:    12345.0,
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},

		//// String TO

		{
			instruction: `dataset:entry | index:"string" | tostring`,
			expected:    "12345",
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:"string" | toboolean`,
			expected:    nil,
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: true,
		},
		{
			instruction: `dataset:entry | index:"string" | toint`,
			expected:    int64(12345),
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:"string" | tofloat`,
			expected:    12345.0,
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},

		//// BOOL TO

		{
			instruction: `dataset:entry | index:"bool" | tostring`,
			expected:    "true",
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:"bool" | toboolean`,
			expected:    true,
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:"bool" | toint`,
			expected:    1,
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:"bool" | tofloat`,
			expected:    1.0,
			data: map[any]any{
				"entry": demo[1],
			},
			hasError: false,
		},

		///

		{
			instruction: `dataset:entry | index:"plz" | toint`,
			expected:    int64(12345),
			data: map[any]any{
				"entry": demo[0],
			},
			hasError: false,
		},

		{
			instruction: `dataset:entry | index:"plz" | toint`,
			expected:    int64(12345),
			data: map[any]any{
				"entry": demo[0],
			},
			hasError: false,
		},

		{
			instruction: `dataset:entry | index:"space-around" | trim`,
			expected:    "12345",
			data: map[any]any{
				"entry": demo[0],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:name | tolower`,
			expected:    "hannes",
			data: map[any]any{
				"entry": demo[0],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:name | toupper`,
			expected:    "HANNES",
			data: map[any]any{
				"entry": demo[0],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:comment | plaintext`,
			expected:    "Test",
			data: map[any]any{
				"entry": demo[0],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:url | urlencoded`,
			expected:    "%24+%26+%3C+%3E+%3F+%3B+%23+%3A+%3D+%2C+%22+%27+~+%2B+%25+%7B+%7D+%7C+%5C+%5E+%5B+%5D+%60",
			data: map[any]any{
				"entry": demo[0],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:url | rawurlencoded`,
			expected:    "%24%20%26%20%3C%20%3E%20%3F%20%3B%20%23%20%3A%20%3D%20%2C%20%22%20%27%20~%20%2B%20%25%20%7B%20%7D%20%7C%20%5C%20%5E%20%5B%20%5D%20%60",
			data: map[any]any{
				"entry": demo[0],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:"url-decoded" | urldecoded`,
			expected:    "$ & < > ? ; # : = , \" ' ~ + % { } | \\ ^ [ ] `",
			data: map[any]any{
				"entry": demo[0],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:name | equal:Hannes:yes:no`,
			expected:    "yes",
			data: map[any]any{
				"entry": demo[0],
			},
			hasError: false,
		},
		{
			instruction: `static: | isnull | not`,
			expected:    true,
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:test | empty:yes:no`,
			expected:    "no",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:test | empty`,
			expected:    false,
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static: | empty`,
			expected:    true,
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static:0 | empty:yes:no`, // empty only checks for empty string
			expected:    nil,
			data:        nil,
			hasError:    true,
		},
		{
			instruction: `static: | empty:yes:no`,
			expected:    "yes",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static: | empty`,
			expected:    true,
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static: | empty:yes`,
			expected:    "yes",
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `static: | isnull`,
			expected:    false, // because is empty string
			data:        map[any]any{},
			hasError:    false,
		},
		{
			instruction: `dataset:entry | index:name | isnull`,
			expected:    false,
			data: map[any]any{
				"entry": demo[0],
			},
			hasError: false,
		},
		{
			instruction: `dataset:entry | index:name2 | isnull`,
			expected:    true,
			data: map[any]any{
				"entry": demo[0],
			},
			hasError: false,
		},
		{
			instruction: `static:"Without Map" | strtolower`,
			expected:    "without map",
			data:        map[any]any{},
			hasError:    false,
		},
	}

	for _, td := range testData {

		t.Run(td.instruction, func(t *testing.T) {

			tt := NewTransformer(&td.data)

			result, err := tt.Transform(td.instruction)
			if td.hasError {
				if err == nil {
					t.Errorf("Expected error, got nil")
				}
				return
			}

			if err != nil {
				t.Error(err)
				return
			}

			if result != td.expected {
				t.Errorf("Expected %v, got %v", td.expected, result)
				t.Errorf("Type of expected: %T", td.expected)
				t.Errorf("Type of result: %T", result)
			}
		})
	}

}
